# Copyright 2024 The SeqIO Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for seqio.metrics."""

from unittest import mock

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np
from seqio import dataset_providers
from seqio import metrics
from seqio import vocabularies


class MetricsTest(absltest.TestCase):

  def test_remove_padding_examples(self):
    mask = np.array([1, 1, 0])
    indices_2d = np.array([[0, 1], [0, 2], [0, 0]])
    model_output = np.array([[1, 2, 3], [1, 3, 4], [1, 2, 3]])
    indices_2d_cleaned, model_output_cleaned = metrics.remove_padding_examples(
        model_output, indices_2d, mask
    )

    np.testing.assert_array_equal(
        indices_2d_cleaned, np.array([[0, 1], [0, 2]])
    )
    np.testing.assert_array_equal(
        model_output_cleaned, np.array([[1, 2, 3], [1, 3, 4]])
    )

  def test_globally_sort_model_output(self):
    indices_2d = np.array([[0, 1], [0, 2], [0, 0]])
    model_output = np.array([[1, 2, 3], [1, 3, 4], [1, 2, 3]])
    model_output_sorted = metrics.globally_sort_model_output(
        model_output, indices_2d
    )

    np.testing.assert_array_equal(
        model_output_sorted, np.array([[1, 2, 3], [1, 2, 3], [1, 3, 4]])
    )


class MetricsComputeTest(parameterized.TestCase):

  @parameterized.parameters(False, True)
  def test_passthrough_legacy_metric(self, do_external_decode: bool):
    mf = lambda targets, predictions: {"accuracy": 1.0}
    metric_obj = metrics.PassthroughLegacyMetric.from_metric_fn(
        mf, None
    ).empty()
    mock_vocab = mock.create_autospec(vocabularies.Vocabulary)

    task_batch = [
        {"targets": [0, 1, 2, 3]},
        {"targets": [1, 2, 3, 4]},
    ]
    model_output = np.array([[0, 1, 2, 3], [1, 2, 3, 4]])
    output_features = {"targets": dataset_providers.Feature(mock_vocab)}

    if do_external_decode:
      model_output = [mock_vocab.decode(tokens) for tokens in model_output]

    metric_instance = metric_obj.from_model_output(
        inputs=task_batch,
        model_output=model_output,
        features=output_features,
    )
    metric_value, _ = metric_instance.actual_compute(
        task_batch, output_features
    )
    self.assertEqual(metric_value["accuracy"], 1.0)


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The SeqIO Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Import to top-level API."""

# pylint:disable=wildcard-import,g-bad-import-order

from seqio.dataset_providers import *
from seqio import evaluation
from seqio import experimental
from seqio.evaluation import Evaluator
from seqio.feature_converters import *
from seqio import helpers
from seqio.loggers import *
from seqio import metrics
from seqio import preprocessors
import seqio.test_utils
from seqio.utils import *
from seqio.vocabularies import *

# Version number.
from seqio.version import __version__



# Copyright 2024 The SeqIO Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for seqio.preprocessors."""

import contextlib
from unittest import mock

from absl.testing import absltest
from seqio import dataset_providers
from seqio import experimental
from seqio import test_utils
from seqio import utils
from seqio import vocabularies
import tensorflow.compat.v2 as tf
import tensorflow_datasets as tfds


assert_dataset = test_utils.assert_dataset
Feature = dataset_providers.Feature
CacheDatasetPlaceholder = dataset_providers.CacheDatasetPlaceholder
MixtureRegistry = dataset_providers.MixtureRegistry
TaskRegistry = dataset_providers.TaskRegistry
ShardInfo = dataset_providers.ShardInfo


class FullyCachedTaskTest(absltest.TestCase):

  def setUp(self):
    super().setUp()

    TaskRegistry.reset()
    MixtureRegistry.reset()

    self.fake_source = dataset_providers.FunctionDataSource(
        lambda split, shuffle_files: tf.data.Dataset.range(2), ['train']
    )

    self.vocabulary = vocabularies.PassThroughVocabulary(100)

    self.metrics_fns = [lambda targets, predictions: 0]

    def fake_preprocessor(ds):
      """Adds one and casts to int32."""
      return ds.map(lambda x: tf.cast(x + 1, tf.int32))

    def fake_preprocessor_of(ds, output_features):
      """Creates output feature dict from scalar input."""
      return ds.map(lambda x: {k: [x] for k in output_features})

    def fake_preprocessor_sl(ds, sequence_length):
      """Concatenates the sequence length to each feature."""
      return ds.map(
          lambda x: {  # pylint:disable=g-long-lambda
              k: tf.concat([v, [sequence_length[k]]], 0) for k, v in x.items()
          }
      )

    def fake_preprocessor_sl_of(ds, sequence_length, output_features):
      """Adds the sequence length to each feature with `add_eos` enabled."""
      return ds.map(
          lambda x: {  # pylint:disable=g-long-lambda
              k: (
                  tf.concat([v, [sequence_length[k]]], 0)
                  if output_features[k].add_eos
                  else v
              )
              for k, v in x.items()
          }
      )

    self.preprocessors = [
        fake_preprocessor,
        fake_preprocessor_of,
        fake_preprocessor_sl,
        fake_preprocessor_sl_of,
    ]

  def validate_fully_cached_task(
      self, name, sequence_length, actual_sequence_length, expected_dataset
  ):
    new_task = TaskRegistry.get(name)
    self.assertLen(new_task.preprocessors, 6)
    self.assertEqual(new_task.metric_fns, self.metrics_fns)
    self.assertIsInstance(new_task.preprocessors[-2], CacheDatasetPlaceholder)
    self.assertTrue(new_task.preprocessors[-2].required)

    with self.assertRaisesWithLiteralMatch(
        ValueError,
        (
            f"Task '{name}' requires caching, but was called with "
            '`use_cached=False`.'
        ),
    ):
      new_task.get_dataset(None)

    # Disable caching restriction to verify dataset is correct.
    new_task.preprocessors[-2]._required = False

    with self.assertRaisesWithLiteralMatch(
        ValueError,
        (
            f"Fully-cached task '{name}' can only be loaded with "
            f'`sequence_length={sequence_length}` or `None`. '
            f'Given sequence_length={actual_sequence_length}.'
        ),
    ):
      new_task.get_dataset(
          {k: v + 1 for k, v in sequence_length.items()}, use_cached=False
      )

    assert_dataset(new_task.get_dataset(None, shuffle=False), expected_dataset)
    assert_dataset(
        new_task.get_dataset(sequence_length, shuffle=False), expected_dataset
    )

  def test_add_fully_cached_task(self):
    preprocessors = list(self.preprocessors)
    preprocessors.insert(2, CacheDatasetPlaceholder())

    TaskRegistry.add(
        'encoder_decoder_task',
        source=self.fake_source,
        preprocessors=preprocessors,
        output_features={
            'inputs': Feature(self.vocabulary, add_eos=True),
            'targets': Feature(self.vocabulary, add_eos=False),
        },
        metric_fns=self.metrics_fns,
    )

    sequence_length = {'inputs': 5, 'targets': 6}
    actual_sequence_length = {'inputs': 6, 'targets': 7}
    experimental.add_fully_cached_task('encoder_decoder_task', sequence_length)
    self.validate_fully_cached_task(
        'encoder_decoder_task_i5_t6',
        sequence_length,
        actual_sequence_length,
        [
            {'inputs': [1, 5, 5], 'targets': [1, 6]},
            {'inputs': [2, 5, 5], 'targets': [2, 6]},
        ],
    )

  def test_add_fully_cached_task_single_feature(self):
    TaskRegistry.add(
        'decoder_task',
        source=self.fake_source,
        preprocessors=self.preprocessors,
        output_features={'targets': Feature(self.vocabulary, add_eos=True)},
        metric_fns=self.metrics_fns,
    )

    sequence_length = {'targets': 6}
    actual_sequence_length = {'targets': 7}
    experimental.add_fully_cached_task('decoder_task', sequence_length)
    self.validate_fully_cached_task(
        'decoder_task_6',
        sequence_length,
        actual_sequence_length,
        [
            {'targets': [1, 6, 6]},
            {'targets': [2, 6, 6]},
        ],
    )

  def test_add_fully_cached_task_unique_prefix(self):
    TaskRegistry.add(
        'feature_prefix_task',
        source=self.fake_source,
        preprocessors=self.preprocessors,
        output_features={
            'tar': Feature(self.vocabulary, add_eos=True),
            'targets': Feature(self.vocabulary, add_eos=False),
        },
        metric_fns=self.metrics_fns,
    )

    sequence_length = {'tar': 5, 'targets': 6}
    actual_sequence_length = {'tar': 6, 'targets': 7}
    experimental.add_fully_cached_task('feature_prefix_task', sequence_length)
    self.validate_fully_cached_task(
        'feature_prefix_task_tar5_targ6',
        sequence_length,
        actual_sequence_length,
        [
            {'tar': [1, 5, 5], 'targets': [1, 6]},
            {'tar': [2, 5, 5], 'targets': [2, 6]},
        ],
    )

  def test_add_fully_cached_task_disallow_shuffling(self):
    TaskRegistry.add(
        'decoder_task',
        source=self.fake_source,
        preprocessors=self.preprocessors,
        output_features={'targets': Feature(self.vocabulary, add_eos=True)},
        metric_fns=self.metrics_fns,
    )

    sequence_length = {'targets': 6}
    new_task = experimental.add_fully_cached_task(
        'decoder_task', sequence_length, disallow_shuffling=True
    )

    # Disable caching restriction to get past cache check.
    new_task.preprocessors[-2]._required = False

    with self.assertRaisesWithLiteralMatch(
        ValueError,
        (
            "Shuffling is disallowed for Task 'decoder_task_6' since its "
            '`shuffle_buffer_size` was set to `None` on construction.'
        ),
    ):
      new_task.get_dataset(None, shuffle=True, use_cached=False)

    new_task.get_dataset(None, shuffle=False, use_cached=False)

  def test_add_fully_cached_mixture(self):
    TaskRegistry.add(
        'task1',
        source=self.fake_source,
        preprocessors=self.preprocessors,
        output_features={'targets': Feature(self.vocabulary, add_eos=False)},
        metric_fns=self.metrics_fns,
    )

    TaskRegistry.add(
        'task2',
        source=self.fake_source,
        preprocessors=self.preprocessors,
        output_features={'targets': Feature(self.vocabulary, add_eos=True)},
        metric_fns=self.metrics_fns,
    )

    MixtureRegistry.add('mix', [('task1', 2), ('task2', lambda x: 1)])

    experimental.add_fully_cached_mixture('mix', sequence_length={'targets': 6})

    new_mix = MixtureRegistry.get('mix_6')
    new_task_names = ('task1_6', 'task2_6')
    self.assertContainsSubset(new_task_names, TaskRegistry.names())

    new_tasks = [TaskRegistry.get(n) for n in new_task_names]

    self.assertCountEqual(new_tasks, new_mix.tasks)
    self.assertEqual(new_mix.get_rate(new_tasks[0]), 2)
    self.assertEqual(new_mix.get_rate(new_tasks[1]), 1)

    with self.assertRaisesWithLiteralMatch(
        ValueError,
        (
            "Task 'task1_6' requires caching, but was called with "
            '`use_cached=False`.'
        ),
    ):
      new_mix.get_dataset(None)

    # Disable caching restriction to get past cache check.
    for t in new_tasks:
      t.preprocessors[-2]._required = False

    with self.assertRaisesWithLiteralMatch(
        ValueError,
        (
            "Fully-cached task 'task1_6' can only be loaded with "
            "`sequence_length={'targets': 6}` or `None`. "
            "Given sequence_length={'targets': 7}."
        ),
    ):
      new_mix.get_dataset({'targets': 7}, use_cached=False)

    expected_dataset = [
        {'targets': [1, 6, 6]},
        {'targets': [2, 6, 6]},
        {'targets': [1, 6]},
        {'targets': [1, 6, 6]},
        {'targets': [2, 6]},
        {'targets': [2, 6, 6]},
    ]

    assert_dataset(
        new_mix.get_dataset(None, shuffle=False).take(6), expected_dataset
    )
    assert_dataset(
        new_mix.get_dataset({'targets': 6}, shuffle=False).take(6),
        expected_dataset,
    )

  def test_add_fully_cached_mixture_disallow_shuffling(self):
    TaskRegistry.add(
        'task1',
        source=self.fake_source,
        preprocessors=self.preprocessors,
        output_features={'targets': Feature(self.vocabulary, add_eos=False)},
        metric_fns=self.metrics_fns,
    )

    TaskRegistry.add(
        'task2',
        source=self.fake_source,
        preprocessors=self.preprocessors,
        output_features={'targets': Feature(self.vocabulary, add_eos=True)},
        metric_fns=self.metrics_fns,
    )

    MixtureRegistry.add('mix', [('task1', 2), ('task2', lambda x: 1)])

    new_mixture = experimental.add_fully_cached_mixture(
        'mix', sequence_length={'targets': 6}, disallow_shuffling=True
    )

    # Disable caching restriction to get past cache check.
    for t in new_mixture.tasks:
      t.preprocessors[-2]._required = False

    with self.assertRaisesWithLiteralMatch(
        ValueError,
        (
            "Shuffling is disallowed for Task 'task1_6' since its "
            '`shuffle_buffer_size` was set to `None` on construction.'
        ),
    ):
      new_mixture.get_dataset(None, shuffle=True, use_cached=False)

    new_mixture.get_dataset(None, shuffle=False, use_cached=False)


class FewshotTest(absltest.TestCase):

  def test_fewshot_data_source(self):
    def fake_dataset_fn(split, shuffle_files, seed=None):
      # Note that for the purposes of this unit test, fake_dataset_fn
      # deliberately does not properly implement shuffling. We test whether
      # FewShotDataSource is robust to this.
      del shuffle_files
      del seed
      return tf.data.Dataset.range(
          *((0, 2) if split == 'validation' else (3, 7))
      )

    # 0 shot
    src = experimental.FewshotDataSource(
        dataset_providers.FunctionDataSource(
            dataset_fn=fake_dataset_fn, splits=['train', 'validation']
        ),
        num_shots=0,
    )
    dataset = src.get_dataset('validation', shuffle=False)
    assert_dataset(
        dataset,
        [
            {
                'eval': 0,
            },
            {'eval': 1},
        ],
    )

    # 1 shot
    preprocessors = [
        utils.map_over_dataset(lambda x: {'inputs': 0, 'targets': x})
    ]
    src = experimental.FewshotDataSource(
        dataset_providers.FunctionDataSource(
            dataset_fn=fake_dataset_fn, splits=['train', 'validation']
        ),
        train_preprocessors=preprocessors,
        eval_preprocessors=preprocessors,
        num_shots=1,
    )

    # When split is 'train', check that 'train' and 'eval' fields of each
    # example are NOT always the same -- this can happen if the underlying
    # dataset_fn does not implement shuffling, causing identical examples from
    # the same split to be zipped together.

    def train_and_eval_fields_always_same(dataset):
      for ex in tfds.as_numpy(dataset):
        if ex['train'] != ex['eval']:
          return False
      return True

    # As long as train and eval fields aren't the same for SOME random seed, we
    # have achieved the desired behavior. We fix the seed for this test because
    # there are some seeds where train and eval fields DO happen to be the same
    # by random chance, which would break this test.
    self.assertFalse(
        train_and_eval_fields_always_same(
            src.get_dataset(split='train', shuffle=True, seed=123)
        )
    )

    # Even when shuffle is off, we don't want the train and eval fields to be
    # the same. Instead, the 'train' field should be deterministically shuffled.
    self.assertFalse(
        train_and_eval_fields_always_same(
            src.get_dataset(split='train', shuffle=False)
        )
    )

    # 3 shot
    src = experimental.FewshotDataSource(
        dataset_providers.FunctionDataSource(
            dataset_fn=fake_dataset_fn, splits=['train', 'validation']
        ),
        train_preprocessors=[
            utils.map_over_dataset(lambda x: {'inputs': 0, 'targets': x})
        ],
        num_shots=3,
    )
    dataset = src.get_dataset('validation', shuffle=False)
    assert_dataset(
        dataset,
        [
            {'eval': 0, 'train': {'inputs': [0, 0, 0], 'targets': [3, 5, 4]}},
            {'eval': 1, 'train': {'inputs': [0, 0, 0], 'targets': [6, 6, 3]}},
        ],
    )
    # Note: the train split has been deterministically shuffled, so the values
    # of the 'targets' field that we test for are deterministic but arbitrary.

    # 3-shot, sharded.
    assert_dataset(
        src.get_dataset(
            'validation', shuffle=False, shard_info=ShardInfo(0, 2)
        ),
        [
            {'eval': 0, 'train': {'inputs': [0, 0, 0], 'targets': [3, 5, 5]}},
        ],
    )
    assert_dataset(
        src.get_dataset(
            'validation', shuffle=False, shard_info=ShardInfo(1, 2)
        ),
        [
            {'eval': 1, 'train': {'inputs': [0, 0, 0], 'targets': [4, 6, 6]}},
        ],
    )
    # Note: the train split has been deterministically shuffled, so the values
    # of the 'targets' field that we test for are deterministic but arbitrary.

    # Missing train
    src = experimental.FewshotDataSource(
        dataset_providers.FunctionDataSource(
            dataset_fn=fake_dataset_fn, splits=['validation']
        ),
        num_shots=3,
    )
    with self.assertRaisesRegex(
        ValueError,
        "Train split 'train' is not one of the original source splits: "
        r'\(\'validation\',\)',
    ):
      dataset = src.get_dataset('validation')

  def test_fewshot_data_source_eval_on_fixed_exemplars(self):
    def fake_dataset_fn(split, shuffle_files, seed=None):
      # Note that for the purposes of this unit test, fake_dataset_fn
      # deliberately does not properly implement shuffling. We test whether
      # FewShotDataSource is robust to this.
      del shuffle_files
      del seed
      return tf.data.Dataset.range(
          *((0, 2) if split == 'validation' else (3, 7))
      )

    # 1 shot
    preprocessors = [
        utils.map_over_dataset(lambda x: {'inputs': 0, 'targets': x})
    ]
    src = experimental.FewshotDataSource(
        dataset_providers.FunctionDataSource(
            dataset_fn=fake_dataset_fn, splits=['train', 'validation']
        ),
        train_preprocessors=preprocessors,
        num_shots=1,
        eval_on_fixed_exemplars=True,
    )

    def exemplars_always_same(dataset):
      """Checks if exemplars are always the same."""
      train_ex = None
      for ex in dataset:
        if train_ex is None:
          train_ex = ex['train']
          continue
        if ex['train'] != train_ex:
          return False
      return True

    # Use 'validation' split for `eval_ds`. Since `train_ds` is initiated from
    # the 'train' split and `eval_on_fixed_exemplars=True`, always use the fixed
    # set of exemplars.
    self.assertTrue(
        exemplars_always_same(
            src.get_dataset(split='validation', shuffle=True, seed=123)
        )
    )

    assert_dataset(
        src.get_dataset('validation', shuffle=False),
        [
            {
                'eval': 0,
                'train': {
                    'inputs': [0],
                    'targets': [3],
                },
            },
            {
                'eval': 1,
                'train': {
                    'inputs': [0],
                    'targets': [3],
                },
            },
        ],
    )

    # `eval_on_fixed_exemplars` is ignored when `split` equals `train_split`.
    self.assertFalse(
        exemplars_always_same(
            src.get_dataset(split='train', shuffle=True, seed=123)
        )
    )

  def test_fewshot_preprocessor(self):
    train_examples = [
        {
            'inputs': 'How many states in the US?',
            'targets': '50',
        },
        {
            'inputs': 'How many cents in a dollar?',
            'targets': '100',
        },
        {
            'inputs': 'How many cents in a quarter?',
            'targets': '25',
        },
    ]

    eval_examples = [
        {
            'inputs': 'Who was in the Beatles?',
            'targets': 'John',
            'answers': ['John', 'Paul', 'George', 'Ringo'],
        },
        {
            'inputs': 'When did the Beatles break up?',
            'targets': '1970',
            'answers': ['1970', 'April 10, 1970', 'April 10', '4/10/1970'],
        },
    ]

    def _from_generator(examples):
      return tf.data.Dataset.from_generator(
          lambda: (x for x in examples),
          output_types={k: tf.string for k in examples[0].keys()},
          output_shapes={
              k: [None] if isinstance(v, list) else []
              for k, v in examples[0].items()
          },
      )

    train_ds = _from_generator(train_examples).repeat()
    eval_ds = _from_generator(eval_examples)

    # 0-shot
    dataset = experimental.fewshot_preprocessor(
        tf.data.Dataset.zip({'eval': eval_ds}),
        inputs_prefix='0 ',
        targets_prefix=' X 1 ',
        example_separator=' X ',
    )
    assert_dataset(
        dataset,
        [
            {
                'inputs': '0 Who was in the Beatles? X 1',
                'targets': 'John',
                'answers': ['John', 'Paul', 'George', 'Ringo'],
            },
            {
                'inputs': '0 When did the Beatles break up? X 1',
                'targets': '1970',
                'answers': ['1970', 'April 10, 1970', 'April 10', '4/10/1970'],
            },
        ],
    )

    # 2-shot
    dataset = experimental.fewshot_preprocessor(
        tf.data.Dataset.zip({'train': train_ds.batch(2), 'eval': eval_ds}),
        inputs_prefix='0 ',
        targets_prefix=' X 1 ',
        example_separator=' X ',
    )
    assert_dataset(
        dataset,
        [
            {
                'inputs': (
                    '0 How many states in the US? X 1 50 X 0 How many cents in '
                    'a dollar? X 1 100 X 0 Who was in the Beatles? X 1'
                ),
                'targets': 'John',
                'answers': ['John', 'Paul', 'George', 'Ringo'],
            },
            {
                'inputs': (
                    '0 How many cents in a quarter? X 1 25 X 0 How many states '
                    'in the US? X 1 50 X 0 When did the Beatles break up? X 1'
                ),
                'targets': '1970',
                'answers': ['1970', 'April 10, 1970', 'April 10', '4/10/1970'],
            },
        ],
    )

    # 1-shot, batched eval
    dataset = experimental.fewshot_preprocessor(
        tf.data.Dataset.zip(
            {'train': train_ds.batch(1), 'eval': eval_ds.batch(2)}
        ),
        inputs_prefix='0 ',
        targets_prefix=' X 1 ',
        example_separator=' X ',
    )
    assert_dataset(
        dataset,
        [
            {
                'inputs': (
                    '0 How many states in the US? X 1 50 X 0 Who was in the '
                    'Beatles? X 1'
                ),
                'targets': 'John',
                'answers': ['John', 'Paul', 'George', 'Ringo'],
            },
            {
                'inputs': (
                    '0 How many states in the US? X 1 50 X 0 When did the '
                    'Beatles break up? X 1'
                ),
                'targets': '1970',
                'answers': ['1970', 'April 10, 1970', 'April 10', '4/10/1970'],
            },
        ],
    )

    # 0-shot + prompt
    dataset = experimental.fewshot_preprocessor(
        tf.data.Dataset.zip({'eval': eval_ds}),
        inputs_prefix='0 ',
        targets_prefix=' X 1 ',
        example_separator=' X ',
        prompt='Answer the question',
    )
    assert_dataset(
        dataset,
        [
            {
                'inputs': 'Answer the question X 0 Who was in the Beatles? X 1',
                'targets': 'John',
                'answers': ['John', 'Paul', 'George', 'Ringo'],
            },
            {
                'inputs': (
                    'Answer the question X 0 When did the Beatles break up? X 1'
                ),
                'targets': '1970',
                'answers': ['1970', 'April 10, 1970', 'April 10', '4/10/1970'],
            },
        ],
    )

    # 2-shot + prompt
    dataset = experimental.fewshot_preprocessor(
        tf.data.Dataset.zip({'train': train_ds.batch(2), 'eval': eval_ds}),
        inputs_prefix='0 ',
        targets_prefix=' X 1 ',
        example_separator=' X ',
        prompt='Answer the question',
    )
    assert_dataset(
        dataset,
        [
            {
                'inputs': (
                    'Answer the question X 0 How many states in the US? X 1 50'
                    ' X 0 How many cents in a dollar? X 1 100 X 0 Who was in'
                    ' the Beatles? X 1'
                ),
                'targets': 'John',
                'answers': ['John', 'Paul', 'George', 'Ringo'],
            },
            {
                'inputs': (
                    'Answer the question X 0 How many cents in a quarter? X 1'
                    ' 25 X 0 How many states in the US? X 1 50 X 0 When did the'
                    ' Beatles break up? X 1'
                ),
                'targets': '1970',
                'answers': ['1970', 'April 10, 1970', 'April 10', '4/10/1970'],
            },
        ],
    )

    # 1-shot, batched eval + prompt
    dataset = experimental.fewshot_preprocessor(
        tf.data.Dataset.zip(
            {'train': train_ds.batch(1), 'eval': eval_ds.batch(2)}
        ),
        inputs_prefix='0 ',
        targets_prefix=' X 1 ',
        example_separator=' X ',
        prompt='Answer the question',
    )
    assert_dataset(
        dataset,
        [
            {
                'inputs': (
                    'Answer the question X 0 How many states in the US? X 1 50'
                    ' X 0 Who was in the Beatles? X 1'
                ),
                'targets': 'John',
                'answers': ['John', 'Paul', 'George', 'Ringo'],
            },
            {
                'inputs': (
                    'Answer the question X 0 How many states in the US? X 1 50'
                    ' X 0 When did the Beatles break up? X 1'
                ),
                'targets': '1970',
                'answers': ['1970', 'April 10, 1970', 'April 10', '4/10/1970'],
            },
        ],
    )

    # 1-shot + eval_on_fixed_exemplars + reverse
    dataset = experimental.fewshot_preprocessor(
        tf.data.Dataset.zip(
            {'train': train_ds.batch(1).take(1).repeat(), 'eval': eval_ds}
        ),
        inputs_prefix='0 ',
        targets_prefix=' X 1 ',
        example_separator=' X ',
        reverse=True,
    )
    assert_dataset(
        dataset,
        [
            {
                'inputs': (
                    '0 How many states in the US? X 1 50 X 0 Who was in the'
                    ' Beatles? X 1'
                ),
                'targets': 'John',
                'answers': ['John', 'Paul', 'George', 'Ringo'],
            },
            {
                'inputs': (
                    '0 How many states in the US? X 1 50 X 0 When did the'
                    ' Beatles break up? X 1'
                ),
                'targets': '1970',
                'answers': ['1970', 'April 10, 1970', 'April 10', '4/10/1970'],
            },
        ],
    )

    # 2-shot + eval_on_fixed_exemplars + reverse
    dataset = experimental.fewshot_preprocessor(
        tf.data.Dataset.zip(
            {'train': train_ds.batch(2).take(1).repeat(), 'eval': eval_ds}
        ),
        inputs_prefix='0 ',
        targets_prefix=' X 1 ',
        example_separator=' X ',
        reverse=True,
    )
    assert_dataset(
        dataset,
        [
            {
                'inputs': (
                    '0 How many cents in a dollar? X 1 100 X 0 How many states '
                    'in the US? X 1 50 X 0 Who was in the Beatles? X 1'
                ),
                'targets': 'John',
                'answers': ['John', 'Paul', 'George', 'Ringo'],
            },
            {
                'inputs': (
                    '0 How many cents in a dollar? X 1 100 X 0 How many states '
                    'in the US? X 1 50 X 0 When did the Beatles break up? X 1'
                ),
                'targets': '1970',
                'answers': ['1970', 'April 10, 1970', 'April 10', '4/10/1970'],
            },
        ],
    )



class SentinelTaskTest(FullyCachedTaskTest):

  def validate_sentinel_task(self, name, sequence_length, expected_dataset):
    new_task = TaskRegistry.get(name)
    # With sentinels inserted we want +1 processors.
    self.assertLen(new_task.preprocessors, 5)
    self.assertEqual(new_task.metric_fns, self.metrics_fns)
    self.assertIsNotNone(new_task.postprocessor)

    assert_dataset(
        new_task.get_dataset(sequence_length, shuffle=False), expected_dataset
    )

  def test_add_sentinel_task(self):
    preprocessors = list(self.preprocessors)

    TaskRegistry.add(
        'encoder_decoder_task',
        source=self.fake_source,
        preprocessors=preprocessors,
        output_features={
            'inputs': Feature(self.vocabulary, add_eos=True),
            'targets': Feature(self.vocabulary, add_eos=False),
        },
        metric_fns=self.metrics_fns,
    )

    sequence_length = {'inputs': 10, 'targets': 11}
    for num_sentinels in [1, 2, 4]:
      experimental.add_task_with_sentinels(
          'encoder_decoder_task', num_sentinels=num_sentinels
      )

    for sentinel_num in [1, 2, 4]:
      sentinel_ids = [
          self.vocabulary.vocab_size - (i + 1) for i in range(sentinel_num)
      ]
      self.validate_sentinel_task(
          f'encoder_decoder_task_{sentinel_num}_sentinel',
          sequence_length,
          [
              {
                  'inputs': [1, 10, 10] + sentinel_ids,
                  'targets': sentinel_ids + [1, 11],
              },
              {
                  'inputs': [2, 10, 10] + sentinel_ids,
                  'targets': sentinel_ids + [2, 11],
              },
          ],
      )


class DisableRegistryTest(absltest.TestCase):

  @contextlib.contextmanager
  def _restore_registry_after_test(self):
    og_tr_add = TaskRegistry.add
    og_tr_add_p = TaskRegistry.add_provider
    og_tr_get = TaskRegistry.get
    og_mr_add = MixtureRegistry.add
    og_mr_add_p = MixtureRegistry.add_provider
    og_mr_get = MixtureRegistry.get
    try:
      yield
    finally:
      TaskRegistry.add = og_tr_add
      TaskRegistry.add_provider = og_tr_add_p
      TaskRegistry.get = og_tr_get
      MixtureRegistry.add = og_mr_add
      MixtureRegistry.add_provider = og_mr_add_p
      MixtureRegistry.get = og_mr_get

  def test_enforce_task_registry_empty(self):
    TaskRegistry.reset()
    MixtureRegistry.reset()
    function_source = dataset_providers.FunctionDataSource(
        dataset_fn=test_utils.get_fake_dataset, splits=['train', 'validation']
    )
    TaskRegistry.add(
        'dummy_task',
        source=function_source,
        output_features=test_utils.FakeTaskTest.DEFAULT_OUTPUT_FEATURES,
    )
    with self.assertRaisesRegex(ValueError, 'The TaskRegistry is non-empty.*'):
      with self._restore_registry_after_test():
        experimental.disable_registry()

  def test_enforce_both_registries_empty(self):
    TaskRegistry.reset()
    MixtureRegistry.reset()
    function_source = dataset_providers.FunctionDataSource(
        dataset_fn=test_utils.get_fake_dataset, splits=['train', 'validation']
    )
    TaskRegistry.add(
        'dummy_task',
        source=function_source,
        output_features=test_utils.FakeTaskTest.DEFAULT_OUTPUT_FEATURES,
    )
    MixtureRegistry.add('dummy_mixture', ['dummy_task'], default_rate=1.0)
    with self.assertRaisesRegex(
        ValueError, 'The TaskRegistry, MixtureRegistry is non-empty.*'
    ):
      with self._restore_registry_after_test():
        experimental.disable_registry()

  def test_task_registry_add_no_op(self):
    TaskRegistry.reset()
    MixtureRegistry.reset()
    with self._restore_registry_after_test():
      experimental.disable_registry()
      function_source = dataset_providers.FunctionDataSource(
          dataset_fn=test_utils.get_fake_dataset, splits=['train', 'validation']
      )
      TaskRegistry.add(
          'dummy_task',
          source=function_source,
          output_features=test_utils.FakeTaskTest.DEFAULT_OUTPUT_FEATURES,
      )
      task = dataset_providers.Task(
          'dummy_task2',
          source=function_source,
          output_features=test_utils.FakeTaskTest.DEFAULT_OUTPUT_FEATURES,
      )
      TaskRegistry.add_provider('dummy_task2', task)
      self.assertEmpty(TaskRegistry.names())

  def test_task_registry_get_error(self):
    TaskRegistry.reset()
    MixtureRegistry.reset()
    with self._restore_registry_after_test():
      experimental.disable_registry()
      with self.assertRaisesRegex(
          ValueError, 'Disabled TaskRegistry.get call.*'
      ):
        TaskRegistry.get('dummy_task')

  def test_mixture_registry_add_no_op(self):
    TaskRegistry.reset()
    MixtureRegistry.reset()
    with self._restore_registry_after_test():
      experimental.disable_registry()
      function_source = dataset_providers.FunctionDataSource(
          dataset_fn=test_utils.get_fake_dataset, splits=['train', 'validation']
      )
      task = dataset_providers.Task(
          'dummy_task',
          source=function_source,
          output_features=test_utils.FakeTaskTest.DEFAULT_OUTPUT_FEATURES,
      )
      TaskRegistry.add_provider('dummy_task2', task)
      MixtureRegistry.add('dummy_mixture', [task], default_rate=1.0)
      mixture = dataset_providers.Mixture('dummy_mixture2', [[task, 1.0]])
      MixtureRegistry.add_provider('dummy_mixture2', mixture)
      self.assertEmpty(TaskRegistry.names())
      self.assertEmpty(MixtureRegistry.names())

  def test_mixture_registry_get_error(self):
    TaskRegistry.reset()
    MixtureRegistry.reset()
    with self.assertRaisesRegex(
        ValueError, 'Disabled MixtureRegistry.get call.*'
    ):
      with self._restore_registry_after_test():
        experimental.disable_registry()
        MixtureRegistry.get('dummy_mixture')



if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The SeqIO Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""SeqIO Beam utilities."""

import functools
import hashlib
import importlib
import json
import operator
import os
from typing import Any, Callable, Dict, Iterable, List, Mapping, Optional, Sequence, Tuple

from absl import logging
import apache_beam as beam
from apache_beam import metrics
import numpy as np
import seqio
import tensorflow.compat.v2 as tf

from array_record.python import array_record_module

PROVENANCE_PREFIX = "provenance/"
TASK_PROVENANCE_KEY = PROVENANCE_PREFIX + "task"
SOURCE_SHARD_PROVENANCE_KEY = PROVENANCE_PREFIX + "source_shard"
SOURCE_SHARD_ID_PROVENANCE_KEY = PROVENANCE_PREFIX + "source_shard_index"
ID_WITHIN_SHARD_PROVENANCE_KEY = PROVENANCE_PREFIX + "index_within_shard"
PREPROCESSORS_SEED_PROVENANCE_KEY = PROVENANCE_PREFIX + "preprocessors_seed"
PROVENANCE_KEYS = [
    TASK_PROVENANCE_KEY,
    SOURCE_SHARD_PROVENANCE_KEY,
    SOURCE_SHARD_ID_PROVENANCE_KEY,
    ID_WITHIN_SHARD_PROVENANCE_KEY,
    PREPROCESSORS_SEED_PROVENANCE_KEY,
]


def _import_modules(modules):
  for module in modules:
    if module:
      importlib.import_module(module)


class PreprocessTask(beam.PTransform):
  """Preprocesses a Task.

  Returns a PCollection of example dicts containing Tensors.
  """

  def __init__(
      self,
      task: seqio.Task,
      split: str,
      *,
      preprocessors_seed: Optional[int] = None,
      setup_fn: Callable[[], None] = lambda: None,
      modules_to_import: Sequence[str] = (),
      add_provenance: bool = False,
      tfds_data_dir: Optional[str] = None,
  ):
    """PreprocessTask constructor.

    Args:
      task: Task, the task to process.
      split: string, the split to process.
      preprocessors_seed: (Optional) int, a seed for stateless random ops in
        task preprocessing.
      setup_fn: (Optional) callable, a function called before loading the task.
      modules_to_import: (Optional) list, modules to import.
      add_provenance: If True, provenance is added to each example.
      tfds_data_dir: (Optional) str, directory where the TFDS datasets are
        stored.

    Raises:
      FileNotFoundError: raised when no shards are found for the task.
    """
    self._task_name = task.name
    self._split = split
    self._preprocessors_seed = preprocessors_seed
    self._setup_fn = setup_fn
    self._modules_to_import = modules_to_import
    self._add_provenance = add_provenance
    self._tfds_data_dir = tfds_data_dir
    self._int64_max = 2**63 - 1
    self.shards = list(enumerate(task.source.list_shards(split)))
    if not self.shards:
      raise FileNotFoundError(f"No shards found for {task.name} {split}")
    logging.info(
        "%s %s %s shards: %d %s",
        task.name,
        split,
        tfds_data_dir,
        len(self.shards),
        ", ".join(["%s" % f[1] for f in self.shards]),
    )

  def _increment_counter(self, name):
    metrics.Metrics.counter(
        str("%s_%s" % (self._task_name, self._split)), name
    ).inc()

  def _emit_examples(self, shard: Tuple[int, str]):
    """Emits examples keyed by shard number and index for a single shard."""
    self._setup_fn()
    _import_modules(self._modules_to_import)
    task = seqio.TaskRegistry.get(self._task_name)

    if self._tfds_data_dir:
      seqio.set_tfds_data_dir_override(self._tfds_data_dir)

    shard_index, shard_name = shard
    logging.info("Processing shard: %s", shard_name)
    self._increment_counter("input-shards")

    # Create a unique, deterministic preprocessors seed for each task and shard.
    md5_digest = hashlib.md5(
        (self._task_name + f"shard{shard_index}").encode()
    ).digest()
    shard_preprocessors_seed = int.from_bytes(md5_digest, "little") + (
        self._preprocessors_seed or 0
    )
    if shard_preprocessors_seed > self._int64_max:
      # The user provided seed is very likely to be much smaller than 2**62,
      # therefore it's safe to just truncated the rest of the bytes and add up.
      md5_digest = md5_digest[:7]
      shard_preprocessors_seed = int.from_bytes(md5_digest, "little") + (
          self._preprocessors_seed or 0
      )
    # Truncate if still a large number.
    shard_preprocessors_seed %= self._int64_max

    ds = task.source.get_dataset(
        split=self._split,
        shard_info=seqio.ShardInfo(
            index=shard_index, num_shards=len(self.shards)
        ),
        shuffle=False,
        seed=shard_preprocessors_seed,
    )
    ds = task.preprocess_precache(ds, seed=shard_preprocessors_seed)
    ds = ds.prefetch(tf.data.AUTOTUNE)

    def _add_provenance(
        index_within_shard: int, ex: Dict[str, Any]
    ) -> Dict[str, Any]:
      ex.update({
          TASK_PROVENANCE_KEY: self._task_name,
          SOURCE_SHARD_PROVENANCE_KEY: shard_name,
          SOURCE_SHARD_ID_PROVENANCE_KEY: shard_index,
          ID_WITHIN_SHARD_PROVENANCE_KEY: index_within_shard,
      })
      if self._preprocessors_seed:
        ex.update({PREPROCESSORS_SEED_PROVENANCE_KEY: self._preprocessors_seed})
      return ex

    for i, ex in enumerate(ds):
      if self._add_provenance:
        ex = _add_provenance(i, ex)
      self._increment_counter("examples")
      # Log every power of two.
      if i & (i - 1) == 0:
        logging.info("Example [%d] = %s", i, ex)
      yield ex

  def expand(self, pipeline):
    return (
        pipeline
        | "create_shards" >> beam.Create(self.shards)
        | "emit_examples" >> beam.FlatMap(self._emit_examples)
    )


class WriteExampleTfRecord(beam.PTransform):
  """Writes examples (dicts) to a TFRecord of tf.Example protos."""

  def __init__(self, output_path: str, num_shards: Optional[int] = None):
    """WriteExampleTfRecord constructor.

    Args:
      output_path: string, path to the output TFRecord file (w/o shard suffix).
      num_shards: (optional) int, number of shards to output or None to use
        liquid sharding.
    """
    self._output_path = output_path
    self._num_shards = num_shards

  def expand(self, pcoll):
    return (
        pcoll
        | beam.Map(seqio.dict_to_tfexample)
        | beam.Reshuffle()
        | beam.io.tfrecordio.WriteToTFRecord(
            self._output_path,
            num_shards=self._num_shards,
            coder=beam.coders.ProtoCoder(tf.train.Example),
        )
    )




class _ArrayRecordSink(beam.io.filebasedsink.FileBasedSink):
  """Sink Class for use in Arrayrecord PTransform."""

  def __init__(
      self,
      file_path_prefix,
      file_name_suffix=None,
      num_shards=0,
      shard_name_template=None,
      coder=beam.coders.coders.ToBytesCoder(),
      compression_type=beam.io.filesystem.CompressionTypes.AUTO,
      preserve_random_access: bool = False,
  ):

    super().__init__(
        file_path_prefix,
        file_name_suffix=file_name_suffix,
        num_shards=num_shards,
        shard_name_template=shard_name_template,
        coder=coder,
        mime_type="application/octet-stream",
        compression_type=compression_type,
    )
    self._preserve_random_access = preserve_random_access

  def open(self, temp_path):
    group_size = 1 if self._preserve_random_access else self.num_shards
    array_writer = array_record_module.ArrayRecordWriter(
        temp_path, f"group_size:{group_size}"
    )
    return array_writer

  def close(self, file_handle):
    file_handle.close()

  def write_encoded_record(self, file_handle, value):
    file_handle.write(value)


class WriteToArrayRecord(beam.PTransform):
  """PTransform for a disk-based write to ArrayRecord."""

  def __init__(
      self,
      file_path_prefix,
      file_name_suffix="",
      num_shards=0,
      shard_name_template=None,
      coder=beam.coders.coders.ToBytesCoder(),
      compression_type=beam.io.filesystem.CompressionTypes.AUTO,
      preserve_random_access: bool = False,
  ):

    self._sink = _ArrayRecordSink(
        file_path_prefix,
        file_name_suffix,
        num_shards,
        shard_name_template,
        coder,
        compression_type,
        preserve_random_access,
    )

  def expand(self, pcoll):
    return pcoll | beam.io.iobase.Write(self._sink)


class WriteExampleArrayRecord(beam.PTransform):
  """Writes examples (dicts) to an ArrayRecord of tf.Example protos."""

  def __init__(
      self,
      output_path: str,
      num_shards: Optional[int] = None,
      preserve_random_access: bool = False,
  ):
    """WriteExampleArrayRecord constructor.

    Args:
      output_path: string, path to the output ArrayRecord file (w/o shard
        suffix).
      num_shards: (optional) int, number of shards to output or None to use
        liquid sharding.
      preserve_random_access: Whether to preserve the random access of the
        written ArrayRecord. If true, set group_size=1, else, set to number of
        shards.
    """
    self._output_path = output_path
    self._num_shards = num_shards
    self._preserve_random_access = preserve_random_access

  def expand(self, pcoll):
    return (
        pcoll
        | beam.Map(seqio.dict_to_tfexample)
        | beam.Reshuffle()
        | WriteToArrayRecord(
            self._output_path,
            num_shards=self._num_shards,
            coder=beam.coders.ProtoCoder(tf.train.Example),
            preserve_random_access=self._preserve_random_access,
        )
    )


class WriteJson(beam.PTransform):
  """Writes datastructures to file as JSON(L)."""

  def __init__(self, output_path: str, prettify: Optional[bool] = True):
    """WriteJson constructor.

    Args:
      output_path: string, path to the output JSON(L) file.
      prettify: bool, whether to write the outputs with sorted keys and
        indentation. Note this not be used if there are multiple records being
        written to the file (JSONL).
    """
    self._output_path = output_path
    self._prettify = prettify

  def _jsonify(self, el):
    if self._prettify:
      return json.dumps(el, sort_keys=True, indent=2)
    else:
      return json.dumps(el)

  def expand(self, pcoll):
    return (
        pcoll
        | beam.Map(self._jsonify)
        | "write_info"
        >> beam.io.WriteToText(
            self._output_path, num_shards=1, shard_name_template=""
        )
    )


class GetInfo(beam.PTransform):
  """Computes info for dataset examples.

  Expects a single PCollections of examples.
  Returns a dictionary with information needed to read the data (number of
  shards, feature shapes and types)
  """

  def __init__(self, num_shards: int, exclude_provenance: bool = True):
    self._num_shards = num_shards
    self._exclude_provenance = exclude_provenance

  def _info_dict(self, ex: List[Dict[str, Any]]):
    if not ex:
      return {}
    assert len(ex) == 1
    ex = ex[0]
    info = {
        "num_shards": self._num_shards,
        "features": {},
        "seqio_version": seqio.__version__,
    }
    feature_dict = info["features"]
    for k, v in ex.items():
      if self._exclude_provenance and k.startswith(PROVENANCE_PREFIX):
        continue
      if isinstance(v, tf.RaggedTensor):
        t = v
      else:
        t = tf.constant(v)
      dtype = t.dtype.name
      shape = t.shape.as_list()
      # Keep all the dimensions but the first if t is not a scalar.
      if shape:
        shape = [None] + shape[1:]
      feature_dict[k] = {"shape": shape, "dtype": dtype}
    return info

  def expand(self, pcoll):
    return (
        pcoll
        | beam.combiners.Sample.FixedSizeGlobally(1)
        | beam.Map(self._info_dict)
    )


class _CountTokens(beam.DoFn):
  """Returns token counts for each feature."""

  def __init__(self, output_features: Mapping[str, seqio.Feature]):
    self._output_features = output_features

  def setup(self):
    # Certain vocabularies are lazy loaded. Since we are running under beam we
    # try to do the loading only once in the setup phase.
    for feat in self._output_features.values():
      v = feat.vocabulary.eos_id
      v = feat.vocabulary.unk_id
      v = feat.vocabulary.pad_id
      del v

  def process(self, ex: Mapping[str, Any]) -> Iterable[Tuple[str, int]]:
    for name, feat in self._output_features.items():
      if (
          name in ex
          and (
              isinstance(ex[name], np.ndarray)
              or isinstance(ex[name], tf.Tensor)
          )
          and ex[name].dtype in (np.int32, np.int64)
      ):
        if isinstance(ex[name], tf.Tensor):
          values = ex[name].numpy()
        else:
          values = ex[name]
        conditions = []
        if feat.vocabulary.eos_id is not None:
          conditions.append((values != feat.vocabulary.eos_id))
        if feat.vocabulary.pad_id is not None:
          conditions.append((values != feat.vocabulary.pad_id))

        if conditions:
          valid_tokens = functools.reduce(operator.and_, conditions)
        else:
          # Assumes all values are valid tokens.
          valid_tokens = np.ones_like(values, dtype=bool)

        num_tokens = int(np.sum(valid_tokens))
        yield (f"{name}_tokens", num_tokens)


class _CountCharacters(beam.DoFn):
  """Returns character counts for each feature.

  This works with both tokenized (integer array) dataset and string dataset. For
  the former, each feature is detokenized and the string length is computed. For
  the latter, the bytes feature is decoded and the string length of that is
  returned.

  Example 1 (tokenized dataset):
  ```python

    Assume that these examples are generated with a vocab that decodes "ea" as
    [4, 5], etc.

    input_examples = [{
        # Decoded as "ea", i.e., length 2 string
        "inputs": np.array([4, 5]),
        # Decoded as "ea test", i.e., length 7 string
        "targets": np.array([4, 5, 10]),
    }, {
        # Decoded as "e", i.e., length 1 string
        "inputs": np.array([4]),
        # Decoded as "asoil", i.e., length 5 string. "1" is an EOS id.
        "targets": np.array([5, 6, 7, 8, 9, 1])
    }]

    This `DoFn` returns (yields each of the 4 elements in sequence):
      [("inputs_chars", 2), ("targets_chars", 7),
       ("inputs_chars", 1), ("targets_chars", 5)]
  ```

  Example 2 (string dataset):
  ```python

    input_examples = [{
        "text": b"this is a string of length 29"
    }, {
        "text": b"this is another string of length 35"
    }]

    This `DoFn` returns (yields each of the 2 elements in sequence):
      [("text_chars", 29), ("text_chars", 35)]
  ```
  """

  def __init__(self, output_features: Mapping[str, seqio.Feature]):
    self._output_features = output_features

  def setup(self):
    # Certain vocabularies are lazy loaded. Since we are running under beam we
    # try to do the loading only once in the setup phase.
    for feat in self._output_features.values():
      v = feat.vocabulary.eos_id
      v = feat.vocabulary.unk_id
      v = feat.vocabulary.pad_id
      del v

  def process(self, ex: Mapping[str, Any]) -> Iterable[Tuple[str, int]]:
    for name, feat in self._output_features.items():
      # We only compute the character length for the rank-1 integer array for
      # the feature using `seqio.SentencePieceVocabulary`.
      if (
          name in ex
          and isinstance(ex[name], np.ndarray)
          and ex[name].dtype in (np.int32, np.int64)
          and feat.rank == 1
          and isinstance(feat.vocabulary, seqio.SentencePieceVocabulary)
      ):
        value = ex[name]
        value = np.abs(value.astype(np.int32))
        decoded = feat.vocabulary.decode_tf(value).numpy().decode("utf-8")

      # If each example in the dataset has the type tf.string, its type
      # becomes `bytes` inside the `ds.as_numpy_iterator()`. This `DoFn` is
      # assumed to be applied to  the examples from such numpy iterator.
      elif name in ex and isinstance(ex[name], bytes):
        decoded = ex[name].decode("utf-8")
      else:
        continue

      yield (f"{name}_chars", len(decoded))


class GetStats(beam.PTransform):
  """Computes statistics for dataset examples.

  The `expand` method expects a PCollection of examples where each example is a
  dictionary of string identifiers (e.g. "inputs" and "targets") mapped to numpy
  array.

  Returns a dictionary with statistics (number of examples, number of tokens)
  prefixed by the identifiers.
  """

  def __init__(
      self,
      output_features: Mapping[str, seqio.Feature],
      task_ids: Optional[Mapping[str, Any]] = None,
      enable_char_counts: bool = False,
  ):
    self._output_features = output_features
    self._task_ids = task_ids or {}
    self._enable_char_counts = enable_char_counts
    logging.info("Getting stats for output features: %s", str(output_features))

  def expand(self, pcoll):
    example_counts = (
        pcoll
        | "count_examples" >> beam.combiners.Count.Globally()
        | "key_example_counts" >> beam.Map(lambda x: ("examples", x))
        | "example_count_dict" >> beam.combiners.ToDict()
    )

    token_counts = pcoll | "count_tokens" >> beam.ParDo(
        _CountTokens(self._output_features)
    )
    total_tokens = (
        token_counts
        | "sum_tokens" >> beam.CombinePerKey(sum)
        | "token_count_dict" >> beam.combiners.ToDict()
    )
    max_tokens = (
        token_counts
        | "max_tokens" >> beam.CombinePerKey(max)
        | "rename_max_stat"
        >> beam.Map(lambda x: (x[0].replace("tokens", "max_tokens"), x[1]))
        | "token_max_dict" >> beam.combiners.ToDict()
    )
    stats = [example_counts, total_tokens, max_tokens]
    if self._enable_char_counts:
      char_length = (
          pcoll
          | beam.ParDo(_CountCharacters(self._output_features))
          | "sum_characters" >> beam.CombinePerKey(sum)
          | "character_length_dict" >> beam.combiners.ToDict()
      )
      stats.append(char_length)

    def _merge_dicts(dicts):
      merged_dict = {}
      for d in dicts:
        assert not set(merged_dict).intersection(d)
        merged_dict.update(d)
      return merged_dict

    if self._task_ids:
      # ids could be Tensors, cast to int.
      self._task_ids = {k: int(v) for k, v in self._task_ids.items()}
      task_ids_dict = {"task_ids": self._task_ids}
      task_ids = (
          pcoll
          | "sample_for_task_ids" >> beam.combiners.Sample.FixedSizeGlobally(1)
          | "create_task_ids" >> beam.Map(lambda _: task_ids_dict)
      )
      stats.append(task_ids)

    return (
        stats
        | "flatten_counts" >> beam.Flatten()
        | "merge_stats" >> beam.CombineGlobally(_merge_dicts)
    )



# Copyright 2024 The SeqIO Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for asserts."""

from absl.testing import absltest
from seqio import dataset_providers
from seqio.test_utils import assert_dataset
from seqio.test_utils import DataInjector
from seqio.test_utils import FakeTaskTest
import tensorflow.compat.v2 as tf

tf.compat.v1.enable_eager_execution()


# Note that the b'string' values are for PY3 to interpret as bytes literals,
# which match the tf.data.Dataset from tensor slices.
class TestUtilsTest(absltest.TestCase):

  def test_assert_dataset(self):
    first_dataset = tf.data.Dataset.from_tensor_slices(
        {'key1': ['val1'], 'key2': ['val2']}
    )
    ragged_tensor_dataset = tf.data.Dataset.from_tensor_slices(
        {'key': tf.ragged.constant([[[1.0, 2.0], [3.0]]])}
    )
    float_dataset = tf.data.Dataset.from_tensor_slices(
        {'float_key1': [0.2, 0.3], 'float_key2': [0.4, 0.5]}
    )

    # Equal
    assert_dataset(first_dataset, {'key1': [b'val1'], 'key2': [b'val2']})
    assert_dataset(
        first_dataset,
        {'key1': [b'val1'], 'key2': [b'val2']},
        expected_dtypes={'key1': tf.string},
    )

    # Equal ragged tensor
    assert_dataset(ragged_tensor_dataset, {'key': [[1.0, 2.0], [3.0]]})

    # Equal floating-point
    assert_dataset(
        float_dataset,
        [
            {'float_key1': 0.2, 'float_key2': 0.4},
            {'float_key1': 0.3, 'float_key2': 0.5},
        ],
    )

    # Close enough floating-point
    assert_dataset(
        float_dataset,
        [
            {'float_key1': 0.20000001, 'float_key2': 0.39999999},
            {'float_key1': 0.30000001, 'float_key2': 0.49999999},
        ],
    )

    # Unequal value
    with self.assertRaises(AssertionError):
      assert_dataset(first_dataset, {'key1': [b'val1'], 'key2': [b'val2x']})

    with self.assertRaises(AssertionError):
      assert_dataset(ragged_tensor_dataset, {'key': [[1.0], [3.0]]})

    # Unequal floating-point, not close enough.
    with self.assertRaises(AssertionError):
      assert_dataset(
          float_dataset,
          [
              {'float_key1': 0.201, 'float_key2': 0.399},
              {'float_key1': 0.301, 'float_key2': 0.499},
          ],
      )

    # Wrong dtype
    with self.assertRaises(AssertionError):
      assert_dataset(
          first_dataset,
          {'key1': [b'val1'], 'key2': [b'val2']},
          expected_dtypes={'key1': tf.int32},
      )

    # Additional key, value
    with self.assertRaises(AssertionError):
      assert_dataset(
          first_dataset,
          {'key1': [b'val1'], 'key2': [b'val2'], 'key3': [b'val3']},
      )

    # Additional key, value
    with self.assertRaises(AssertionError):
      assert_dataset(
          first_dataset,
          {'key1': [b'val1'], 'key2': [b'val2'], 'key3': [b'val3']},
      )


class TasksTest(FakeTaskTest):

  def test_data_injection(self):
    def ds_fn(split, shuffle_files):
      del shuffle_files
      data = {'train': {'data': b'not used'}}
      ds = tf.data.Dataset.from_tensors(data[split])
      return ds

    source = dataset_providers.FunctionDataSource(
        dataset_fn=ds_fn, splits=['train']
    )

    dataset_providers.TaskRegistry.add(
        'test_data_injection_task',
        source=source,
        preprocessors=[],
        output_features={},
        metric_fns=[],
    )

    data = {'train': {'data': b'This data is not used.'}}
    with DataInjector('test_data_injection_task', data):
      pass

    task = dataset_providers.TaskRegistry.get('test_data_injection_task')
    self.assertIs(source, task._source)


if __name__ == '__main__':
  absltest.main()
